"""Test querying with properties."""
import pytest
from .fixtures import fixture_db_driver
from reasoner_transpiler.attributes import set_custom_attribute_types, set_custom_attribute_skip_list, \
    reset_custom_attribute_types, DEFAULT_ATTRIBUTE_TYPE
from reasoner_transpiler.cypher import get_query


def test_numeric(db_driver):
    """Test querying with numeric property."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Gene",
                "constraints": [
                    {"id": "length",
                     "value": 277,
                     "operator": "==="}
                ]
            },
        },
        "edges": {},
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1

    node_1 = list(output["knowledge_graph"]["nodes"].values())[0]
    assert node_1["name"] == "CASP3"
    assert "length" not in node_1
    assert "element_id" not in node_1


def test_string(db_driver):
    """Test querying with string property."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Gene",
                "constraints": [
                    {"id": "chromosome",
                     "value": "17",
                     "operator": "==="}
                ]
            },
        },
        "edges": {},
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1
    results = sorted(
        output["knowledge_graph"]["nodes"].values(),
        key=lambda node: node["name"],
    )
    expected_nodes = [
        "BRCA1",
    ]
    for ind, result in enumerate(results):
        assert result["name"] == expected_nodes[ind]


def test_bool(db_driver):
    """Test querying with boolean property."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:ChemicalSubstance",
            },
            "n1": {
                "categories": "biolink:Disease",
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "predicates": "biolink:treats",
                "attribute_constraints": [
                    {"id": "fda_approved",
                     "value": True,
                     "operator": "==="}
                ]
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1
    results = sorted(
        output["knowledge_graph"]["nodes"].values(),
        key=lambda node: node["name"],
    )
    expected_nodes = [
        "metformin", "type 2 diabetes mellitus",
    ]
    for ind, result in enumerate(results):
        assert result["name"] == expected_nodes[ind]


def test_publications(db_driver):
    """Test publications."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["NCBIGene:836"],
            },
            "n1": {
                "ids": ["NCBIGene:841"],
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph,dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    edges = output["knowledge_graph"]["edges"]
    assert len(edges) == 1
    attributes = list(edges.values())[0]["attributes"]
    assert any([attribute == {
        "original_attribute_name": "publications",
        "attribute_type_id": "biolink:publications",
        "value": ["xxx"],
        "value_type_id": "linkml:Uriorcurie"
    } for attribute in attributes])


def test_empty_constraints(db_driver):
    """Test querying with empty 'constraints' property."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Gene",
                "constraints": [],
            },
            "n1": {
                "constraints": [],
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "attribute_constraints": [],
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 10


def test_valid_biolink_attribute_without_mapping(db_driver):
    """publications is included in the default attribute types mapping included in the codebase,
    but here we test that it still gets recognized as being biolink compliant when it's not in the mapping"""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["MONDO:0005148"],
            },
            "n1": {
                "ids": ["NCBIGene:841"],
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    edges = output["knowledge_graph"]["edges"]
    assert len(edges) == 1
    edge = list(edges.values())[0]
    assert "p_value" not in edge
    attributes = edge["attributes"]
    assert any([attribute == {
        "original_attribute_name": "p_value",
        "attribute_type_id": "biolink:p_value",
        "value": 0.000007
    } for attribute in attributes])
    reset_custom_attribute_types()


def test_invalid_biolink_attribute_without_mapping(db_driver):
    """Test to make sure an attribute that isn't biolink compliant, and isn't custom mapped, returns correctly."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["NCBIGene:836"]
            },
            "n1": {
                "ids": ["MONDO:0005148"]
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1"
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    edges = output["knowledge_graph"]["edges"]
    attributes = list(edges.values())[0]["attributes"]
    expected_attribute = {
        "original_attribute_name": "non_biolink_attribute",
        "attribute_type_id": DEFAULT_ATTRIBUTE_TYPE['attribute_type_id'],
        "value": "xxx123"
    }
    assert any([attribute == expected_attribute for attribute in attributes])


def test_json_attributes(db_driver):
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["NCBIGene:672"]
            },
            "n1": {
                "ids": ["MONDO:0004993"]
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "predicates": "biolink:gene_associated_with_condition"
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph,dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    edges = output["knowledge_graph"]["edges"]
    assert len(edges) == 1
    correct_attributes = 0
    attributes = list(edges.values())[0]["attributes"]
    for attribute in attributes:
        if attribute.get("original_attribute_name", None) == "publications":
            assert attribute["value"] == ["xxx"]
            correct_attributes += 1
        elif attribute["attribute_type_id"] == "json_attribute_1":
            assert attribute["value"] == "json_value_1"
            correct_attributes += 1
        elif attribute["attribute_type_id"] == "json_attribute_2":
            correct_attributes += 1
        elif attribute["attribute_type_id"] == "json_attribute_3":
            assert attribute["value"] == "json_value_3"
            nested_attribute = attribute["attributes"][0]
            assert nested_attribute["attribute_type_id"] == "nested_json_attribute_1"
            assert nested_attribute["value"] == "nested_json_value_1"
            correct_attributes += 1
    assert correct_attributes == 4


def test_props_customization(db_driver):
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["NCBIGene:836"],
            },
            "n1": {
                "ids": ["NCBIGene:841"],
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
        },
    }

    # put publications in the skip list and make sure it doesn't get included
    set_custom_attribute_skip_list(['publications'])
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    edges = output["knowledge_graph"]["edges"]
    assert len(edges) == 1
    attributes = list(edges.values())[0].get("attributes", [])
    assert not any([attribute["attribute_type_id"] == "biolink:publications" for attribute in attributes])
    assert not any([attribute["attribute_type_id"] == "publications" for attribute in attributes])

    # reset the skip list
    set_custom_attribute_skip_list([])

    # set custom attribute and value types for publications and make sure they get used instead of biolink or defaults
    set_custom_attribute_types({
        "publications": {
            "attribute_type_id": "transpiler:custom_attribute_type",
            "value_type_id": "transpiler:custom_value_type"
        }
    })
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    edges = output["knowledge_graph"]["edges"]
    assert len(edges) == 1
    attributes = list(edges.values())[0]["attributes"]
    assert any([attribute == {
        "original_attribute_name": "publications",
        "attribute_type_id": "transpiler:custom_attribute_type",
        "value": ["xxx"],
        "value_type_id": "transpiler:custom_value_type"
    } for attribute in attributes])
    reset_custom_attribute_types()



import os
from bmt import Toolkit

from bmt.toolkit import ClassDefinition
from bmt.toolkit import SlotDefinition

BIOLINK_MODEL_VERSION = os.environ.get('BL_VERSION', '4.2.6-rc5')
BIOLINK_MODEL_SCHEMA_URL = f"https://raw.githubusercontent.com/biolink/biolink-model/v{BIOLINK_MODEL_VERSION}/biolink-model.yaml"
PREDICATE_MAP_URL = f"https://raw.githubusercontent.com/biolink/biolink-model/v{BIOLINK_MODEL_VERSION}/predicate_mapping.yaml"

bmt = Toolkit(schema=BIOLINK_MODEL_SCHEMA_URL, predicate_map=PREDICATE_MAP_URL)
ALL_BIOLINK_ENUMS = bmt.view.all_enums().keys()


def is_biolink_slot(obj):
    return isinstance(obj, SlotDefinition)


def get_slot_uri(obj):
    return obj["slot_uri"]


def is_biolink_class(obj):
    return isinstance(obj, ClassDefinition)


def get_class_uri(obj):
    return obj["class_uri"]



"""Test invalid query graphs."""
import pytest

from reasoner_transpiler.cypher import get_query
from reasoner_transpiler.exceptions import InvalidPredicateError, InvalidQualifierError, \
    InvalidQualifierValueError, UnsupportedError


def test_invalid_node():
    """Test that an invalid node property value throws an error."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:BiologicalEntity",
                "constraints": [
                    {"id": "test:invalid_constraint",
                     "value": {"a": 1}}
                ]
            },
        },
        "edges": dict(),
    }
    with pytest.raises(UnsupportedError):
        get_query(qgraph)


def test_invalid_predicate():
    """Test that an invalid edge predicate throws an error."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["MONDO:0005148"],
            },
            "n1": {
                "categories": ["biolink:PhenotypicFeature"],
            },
        },
        "edges": {
            "e0": {
                "subject": "n0",
                "object": "n1",
                "predicates": ["biolink:invalid_predicate"],
            },
        },
    }
    with pytest.raises(InvalidPredicateError):
        query = get_query(qgraph)

    """Test that an invalid edge predicate throws an error, along with a valid predicate."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["MONDO:0005148"],
            },
            "n1": {
                "categories": ["biolink:PhenotypicFeature"],
            },
        },
        "edges": {
            "e0": {
                "subject": "n0",
                "object": "n1",
                "predicates": ["biolink:invalid_predicate", "biolink:associated_with"],
            },
        },
    }
    with pytest.raises(InvalidPredicateError):
        query = get_query(qgraph)


def test_invalid_qualifier():
    """Test that an invalid edge qualifier throws an error."""
    qgraph = {
        "nodes": {
            "n0": {},
            "n1": {
                "ids": ["NCBIGene:283871"]
            },
        },
        "edges": {
            "e0": {
                "subject": "n0",
                "object": "n1",
                "predicates": ["biolink:affects"],
                "qualifier_constraints": [
                    {
                        "qualifier_set": [
                            {
                                "qualifier_type_id": "bogus_qualifier_1",
                                "qualifier_value": "abundance"
                            },
                        ]
                    },
                ]
            },
        },
    }
    with pytest.raises(InvalidQualifierError):
        query = get_query(qgraph)


def test_invalid_qualifier_value():
    """Test that an invalid edge qualifier value throws an error."""
    qgraph = {
        "nodes": {
            "n0": {},
            "n1": {
                "ids": ["NCBIGene:283871"]
            },
        },
        "edges": {
            "e0": {
                "subject": "n0",
                "object": "n1",
                "predicates": ["biolink:affects"],
                "qualifier_constraints": [
                    {
                        "qualifier_set": [
                            {
                                "qualifier_type_id": "object_aspect_qualifier",
                                "qualifier_value": "bogus_value"
                            },
                        ]
                    },
                ]
            },
        },
    }
    with pytest.raises(InvalidQualifierValueError):
        query = get_query(qgraph)



import json
import os
from pathlib import Path

from .biolink import bmt, is_biolink_slot, is_biolink_class, get_slot_uri, get_class_uri

DIR_PATH = Path(__file__).parent


def get_attribute_types_from_config():
    with open(DIR_PATH / "attribute_types.json", "r") as stream:
        return json.load(stream)


ATTRIBUTE_TYPES = get_attribute_types_from_config()

DEFAULT_ATTRIBUTE_TYPE = {
    "attribute_type_id": "biolink:Attribute"
}

ATTRIBUTE_SKIP_LIST = []

PRIMARY_KNOWLEDGE_SOURCE = "primary_knowledge_source"
AGGREGATOR_KNOWLEDGE_SOURCE = "aggregator_knowledge_source"

PROVENANCE_TAG = os.environ.get('PROVENANCE_TAG', 'reasoner-transpiler')


# This function takes EDGE_SOURCE_PROPS properties from results, converts them into proper
# TRAPI dictionaries, and assigns the proper upstream ids to each resource. It does not currently attempt to avoid
# duplicate aggregator results, which shouldn't exist in the graphs.
def construct_sources_tree(primary_knowledge_source, aggregator_knowledge_sources):

    if not primary_knowledge_source:
        return [{"resource_id": PROVENANCE_TAG,
                 "resource_role": "primary_knowledge_source"}]

    # set the primary knowledge source
    formatted_sources = [{"resource_id": primary_knowledge_source,
                          "resource_role": "primary_knowledge_source"}]

    # walk through the aggregator lists and construct the chains of provenance
    terminal_aggregators = set()
    for aggregator_list in aggregator_knowledge_sources:
        # each aggregator list should be in order, so we can deduce the upstream chains
        last_aggregator = None
        for aggregator_knowledge_source in aggregator_list:
            formatted_sources.append({
                "resource_id": aggregator_knowledge_source,
                "resource_role": "aggregator_knowledge_source",
                "upstream_resource_ids": [last_aggregator] if last_aggregator else [primary_knowledge_source]
            })
            last_aggregator = aggregator_knowledge_source
        # store the last aggregator in the list, because this will be an upstream source for the plater one
        terminal_aggregators.add(last_aggregator)
    # add PROVENANCE_TAG as the most downstream aggregator,
    # it will have as upstream either the primary ks or all of the furthest downstream aggregators if they exist
    # this will be used by applications like Plater which need to append themselves as an aggregator
    formatted_sources.append({
        "resource_id": PROVENANCE_TAG,
        "resource_role": "aggregator_knowledge_source",
        "upstream_resource_ids": list(terminal_aggregators) if terminal_aggregators else [primary_knowledge_source]
    })
    return formatted_sources


def transform_attributes(result_entity, node=False):

    # construct a valid TRAPI entity to return in trapi_entity
    trapi_entity = {}

    for attribute in ATTRIBUTE_SKIP_LIST:
        result_entity.pop(attribute, None)

    # an "attributes" attribute in neo4j should be a list of json strings,
    # attempt to start the attributes section of transformed attributes with its contents,
    # here we are assuming the attributes in "attributes" are already valid trapi
    json_attributes = []
    json_attributes_attribute = result_entity.pop('attributes', None)
    if json_attributes_attribute:
        if isinstance(json_attributes_attribute, list):
            try:
                json_attributes = [json.loads(json_attribute_string)
                                   for json_attribute_string in json_attributes_attribute]
            except json.JSONDecodeError:
                print(f'!!! JSONDecodeError while parsing attributes property, ignoring: {json_attributes_attribute}')
        else:
            print(f'!!! the attributes edge property should be a list, ignoring: {json_attributes_attribute}')
        trapi_attributes = json_attributes
    else:
        trapi_attributes = []

    # if it's an edge handle provenance (sources) and qualifiers
    if not node:
        # extract properties for provenance, construct the sources section
        primary_knowledge_source = result_entity.pop(PRIMARY_KNOWLEDGE_SOURCE, None)
        # get any properties that start with AGGREGATOR_KNOWLEDGE_SOURCE, this handles the possibility of edges
        # with multiple aggregator knowledge source lists like aggregator_knowledge_source_2
        aggregator_knowledge_source_keys = [ks_attribute for ks_attribute in result_entity.keys()
                                            if ks_attribute.startswith(AGGREGATOR_KNOWLEDGE_SOURCE)]
        aggregator_knowledge_sources = [result_entity[key] for key in aggregator_knowledge_source_keys]
        for ks_property in aggregator_knowledge_source_keys:
            result_entity.pop(ks_property)
        trapi_entity["sources"] = construct_sources_tree(primary_knowledge_source, aggregator_knowledge_sources)

        # find and format attributes that are qualifiers
        qualifiers = [key for key in result_entity if bmt.is_qualifier(key)]
        if qualifiers:
            trapi_entity["qualifiers"] = [{"qualifier_type_id": f"biolink:{qualifier}",
                                          "qualifier_value": result_entity.pop(qualifier)}
                                          for qualifier in qualifiers]

    # for attributes that aren't in ATTRIBUTE_TYPES, see if they are valid biolink attributes
    # TODO this is still inefficient - we should really map all the possible attributes on start up
    for attribute in result_entity:
        if attribute not in ATTRIBUTE_TYPES:
            bmt_element = bmt.get_element(attribute)
            if not bmt_element:
                ATTRIBUTE_TYPES[attribute] = DEFAULT_ATTRIBUTE_TYPE
            else:
                # This looks in the biolink model for the slot_uri or class_uri depending on if the element
                # is a slot or a class and attempts to populate the attribute_type_id and value_type_id with something
                # useful. Technically classes probably shouldn't be included as attribute_type_id but examples exist
                # and it seems better than having a default value that also isn't compliant.
                attribute_type_id = f'biolink:{attribute}'
                value_type_id = None
                if is_biolink_slot(bmt_element):
                    bl_slot_uri = get_slot_uri(bmt_element)
                    value_type_id = bl_slot_uri if bl_slot_uri != attribute_type_id else None
                elif is_biolink_class(bmt_element):
                    attribute_type_id = get_class_uri(bmt_element)
                attribute_mapping = {
                    'attribute_type_id': attribute_type_id,
                    **({'value_type_id': value_type_id} if value_type_id else {})
                }
                # add it to ATTRIBUTE_TYPES, so we don't need to check biolink again
                ATTRIBUTE_TYPES[attribute] = attribute_mapping

    # format the rest of the attributes, look up their attribute type and value type
    trapi_attributes.extend([
        {'original_attribute_name': key,
         'value': value,
         # ATTRIBUTE_TYPES is a mapping for things like attribute_type_id, value_type_id or other TRAPI fields
         **ATTRIBUTE_TYPES.get(key)}
        for key, value in result_entity.items()
    ])
    if trapi_attributes:
        trapi_entity["attributes"] = trapi_attributes
    return trapi_entity


def set_custom_attribute_types(attribute_types: dict):
    global ATTRIBUTE_TYPES
    ATTRIBUTE_TYPES = attribute_types


def set_custom_attribute_skip_list(skip_list: list):
    global ATTRIBUTE_SKIP_LIST
    ATTRIBUTE_SKIP_LIST = skip_list


def reset_custom_attribute_types():
    global ATTRIBUTE_TYPES
    ATTRIBUTE_TYPES = get_attribute_types_from_config()




"""Test transpiler edge cases."""
from reasoner_transpiler.cypher import get_query
from .fixtures import fixture_db_driver
#import pytest


def test_categories(db_driver):
    """Test multiple categories."""
    qgraph = {
        "nodes": {"n0": {"categories": [
            "biolink:Disease",
            "biolink:Gene",
        ]}},
        "edges": dict(),
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 12


def test_empty(db_driver):
    """Test empty qgraph."""
    qgraph = {
        "nodes": dict(),
        "edges": dict(),
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 0
    assert output["knowledge_graph"]["nodes"] == {}
    assert output["knowledge_graph"]["edges"] == {}


def test_category_not_specified(db_driver):
    """Test node with type None."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["MONDO:0014488"]
            }
        },
        "edges": dict(),
    }
    dialect, driver = db_driver
    cypher = get_query(qgraph, dialect=dialect)
    assert "NamedThing" in cypher
    output = driver.run(cypher, convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1


def test_category_none(db_driver):
    """Test node with type None."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["MONDO:0014488"],
                "categories": None
            }
        },
        "edges": dict(),
    }
    dialect, driver = db_driver
    cypher = get_query(qgraph, dialect=dialect)
    assert "NamedThing" in cypher
    output = driver.run(cypher, convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1


def test_relation_none(db_driver):
    """Test edge with relation None."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Disease",
            },
            "n1": {
                "categories": "biolink:Gene",
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "relation": None,
            }
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 5


def test_qnode_addl_null(db_driver):
    """Test qnode with null-valued additional property."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Disease",
            },
            "n1": {
                "categories": "biolink:Gene",
                "chromosome": None,
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            }
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph,dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 5


def test_predicate_none(db_driver):
    """Test edge with predicate None."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Disease",
            },
            "n1": {
                "categories": "biolink:Gene",
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "predicates": None,
            }
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph,dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 5


def test_fancy_key(db_driver):
    """Test qnode/qedge keys with unusual characters."""
    qgraph = {
        "nodes": {
            "type-2 diabetes": {
                "categories": "biolink:Disease",
            },
            "n1": {
                "categories": "biolink:Gene",
            },
        },
        "edges": {
            "interacts with": {
                "subject": "type-2 diabetes",
                "object": "n1",
            }
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 5


#This one kills memgraph 3.3.0.  See https://github.com/memgraph/memgraph/issues/3080
def test_backwards_predicate(db_driver):
    """Test an extra backwards predicate."""
    qgraph = {
        "nodes": {
            "type-2 diabetes": {
                "ids": ["MONDO:0005148"],
                "categories": "biolink:Disease",
            },
            "drug": {
                "categories": "biolink:ChemicalSubstance",
            },
        },
        "edges": {
            "related to": {
                "subject": "type-2 diabetes",
                "object": "drug",
                "predicates": ["biolink:related_to", "biolink:treats"]
            }
        }
    }
    dialect, driver = db_driver
    query = get_query(qgraph, dialect=dialect)
    print(query)
    output = driver.run(query, convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 3


def test_index_usage_empty_labels():
    """
    Test when using single labels, checks if id index is with the node type is used
    """
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["MONDO:0005148"],
                "categories": [],
            }
        },
        "edges": {}
    }
    cypher = get_query(qgraph, **{"use_hints": True})
    # superclass node_id is suffixed with _superclass
    assert "USING INDEX `n0`:`biolink:NamedThing`(id)" in cypher


def test_index_usage_missing_labels():
    """
    Test when using single labels, checks if id index is with the node type is used
    """
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["MONDO:0005148"]
            }
        },
        "edges": {}
    }
    cypher = get_query(qgraph, **{"use_hints": True})
    # superclass node_id is suffixed with _superclass
    assert "USING INDEX `n0`:`biolink:NamedThing`(id)" in cypher


def test_index_usage_single_labels():
    """
    Test when using single labels, checks if id index is with the node type is used
    """
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["MONDO:0005148"],
                "categories": "biolink:Disease",
            }
        },
        "edges": {}
    }
    cypher = get_query(qgraph, **{"use_hints": True})
    # superclass node_id is suffixed with _superclass
    assert "USING INDEX `n0`:`biolink:Disease`(id)" in cypher


def test_index_usage_multiple_labels():
    """
    When multiple labels are used `biolink:NamedThing` index to be used
    """
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["MONDO:0005148"],
                "categories": ["biolink:Disease", "biolink:PhenotypicFeature"]
            }
        },
        "edges": {}
    }
    cypher = get_query(qgraph, **{"use_hints": True})
    # superclass node_id is suffixed with _superclass
    assert "USING INDEX `n0`:`biolink:NamedThing`(id)" in cypher


def test_index_usage_with_subclass():
    """Test an extra backwards predicate."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["MONDO:0005148"],
                "categories": "biolink:Disease",
            },
            "n1": {
                "categories": "biolink:ChemicalSubstance",
            },
        },
        "edges": {
            "related to": {
                "subject": "type-2 diabetes",
                "object": "drug",
                "predicates": ["biolink:related_to", "biolink:treats"]
            }
        }
    }
    cypher = get_query(qgraph, **{"use_hints": True})
    # superclass node_id is suffixed with _superclass
    assert "USING INDEX `n0_superclass`:`biolink:Disease`(id)" in cypher



"""Test casing."""
import pytest

from reasoner_transpiler.util import pascal_case, snake_case, space_case


def test_space():
    """Test conversion to space case."""
    assert space_case("ChemicalSubstance") == "chemical substance"
    assert space_case([
        "ChemicalSubstance",
        "biological_process"
    ]) == [
        "chemical substance",
        "biological process",
    ]
    with pytest.raises(ValueError):
        space_case({"a": "ChemicalSubstance"})


def test_snake():
    """Test conversion to snake_case."""
    assert snake_case("ChemicalSubstance") == "chemical_substance"
    assert snake_case([
        "ChemicalSubstance",
        "Biological Process"
    ]) == [
        "chemical_substance",
        "biological_process",
    ]
    with pytest.raises(ValueError):
        snake_case({"a": "ChemicalSubstance"})


def test_pascal():
    """Test conversion to PascalCase."""
    assert pascal_case("chemical_substance") == "ChemicalSubstance"
    assert pascal_case([
        "chemical_substance",
        "biological process",
    ]) == [
        "ChemicalSubstance",
        "BiologicalProcess"
    ]
    with pytest.raises(ValueError):
        pascal_case({"a": "ChemicalSubstance"})



"""Test entity subclassing."""
import copy
import pytest
import json
from reasoner_transpiler.cypher import get_query
from .fixtures import fixture_db_driver


def get_node_binding_ids_from_neo4j_driver_output(output):
    return [binding["id"] for result in output["results"]
            for binding_list in result["node_bindings"].values()
            for binding in binding_list]


def test_node_subclass(db_driver):
    """Test node-only subclass query."""
    # for queries without edges, we don't include subclasses of the query node or subclass edges
    qgraph = {
        "nodes": {"n0": {"ids": ["MONDO:0000001"]}},
        "edges": dict(),
    }
    original_qgraph = copy.deepcopy(qgraph)
    dialect, driver = db_driver
    query = get_query(qgraph, dialect=dialect)
    # assert qgraph != original_qgraph  # subclass queries should change the qgraph and add qnodes etc
    output = driver.run(query, convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 1
    assert len(output['knowledge_graph']['nodes']) == 1
    assert len(output['knowledge_graph']['edges']) == 0

    node_binding_ids = get_node_binding_ids_from_neo4j_driver_output(output)
    assert 'MONDO:0015967' not in node_binding_ids
    assert 'MONDO:0005148' not in node_binding_ids

def jsort(j):
    """Sort JSON object."""
    isorted = json.loads(json.dumps(j, sort_keys=True))
    #This takes care of the dicts, but there's still potentially a list of dicts that needs to be sorted by the value of their id
    isorted["analyses"][0]["edge_bindings"]["e01"]=sorted(isorted["analyses"][0]["edge_bindings"]["e01"],key=lambda x:x["id"])
    return isorted

def test_onehop_subclass(db_driver):
    """Test one-hop subclass query."""
    qgraph = {
        "nodes": {
            "n0": {"ids": ["MONDO:0000001"]},
            "n1": {},
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 12
    assert len(output['auxiliary_graphs']) == 11
    expected_result = jsort({'analyses': [{'edge_bindings': {'e01': [{'id': 't2d_invalid_predicate_albuminaria_t2d_isa_disease', 'attributes': []}, {'id': 't2d_has_phenotype_albuminaria_t2d_isa_disease', 'attributes': []}]}, 'resource_id': 'reasoner-transpiler'}], 'node_bindings': {'n0': [{'id': 'MONDO:0000001', 'attributes': []}], 'n1': [{'id': 'HP:0012592', 'attributes': []}]}})
    assert any(expected_result == jsort(result) for result in output['results'])
    assert 'aux_t2d_has_phenotype_albuminaria_t2d_isa_disease' in output['auxiliary_graphs']
    assert 't2d_has_phenotype_albuminaria_t2d_isa_disease' in output['knowledge_graph']['edges']
    assert {"attribute_type_id": "biolink:support_graphs", "value": ["aux_anagliptin_treats_t2d_t2d_isa_disease"]}\
           in output['knowledge_graph']['edges']['anagliptin_treats_t2d_t2d_isa_disease']['attributes']


def test_onehop_subclass_categories():
    """Test one-hop subclass query."""
    qgraph = {
        "nodes": {
            "n0": {"ids": ["HP:0011015"], "categories": ["biolink:PhenotypicFeature"]},
            "n1": {},
        },
        "edges": {
            "e01": {
                "subject": "n1",
                "object": "n0",
                "predicates": ["biolink:treats"]
            },
        },
    }
    query = get_query(qgraph)
    #make sure that the class (PhenotypicFeature) has been removed from n0
    clause = query.split('WITH')[0]
    elements = clause.split('-')
    checked = False
    for element in elements:
        if '`n0`' in element:
            checked = True
            assert 'PhenotypicFeature' not in element
    assert checked


def test_backward_subclass(db_driver):
    """Test pinned-object one-hop subclass query."""
    qgraph = {
        "nodes": {
            "n0": {"ids": ["MONDO:0000001"]},
            "n1": {},
        },
        "edges": {
            "e01": {
                "subject": "n1",
                "object": "n0",
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 12
    assert len(output['auxiliary_graphs']) == 11


def test_pinned_subclass(db_driver):
    """Test both-pinned subclass query."""
    qgraph = {
        "nodes": {
            "n0": {"ids": ["MONDO:0000001"]},
            "n1": {"ids": ["HP:0000118"]},
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 1
    assert output["results"][0]["node_bindings"] == {
        "n0": [{"id": "MONDO:0000001", "attributes": []}],
        "n1": [{"id": "HP:0000118", "attributes": []}],
    }


def test_same_pinned_subclass(db_driver):
    """Test both-pinned subclass query."""
    qgraph = {
        "nodes": {
            "n0": {"ids": ["MONDO:0000001"]},
            "n1": {"ids": ["MONDO:0000001"]},
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 1
    assert len(output['knowledge_graph']['nodes']) == 3
    assert len(output['knowledge_graph']['edges']) == 4
    assert len(output['auxiliary_graphs']) == 2


def test_multihop_subclass(db_driver):
    """Test multi-hop subclass query."""
    qgraph = {
        "nodes": {
            "n0": {"ids": ["MONDO:0000001"]},
            "n1": {},
            "n2": {},
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
            "e12": {
                "subject": "n1",
                "object": "n2",
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 35


def test_dont_subclass(db_driver):
    """Test disallowing subclassing."""
    qgraph = {
        "nodes": {"n0": {"ids": ["MONDO:0000001"]}},
        "edges": dict(),
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect, subclass=False), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 1
    assert output["results"][0]["node_bindings"] == {
        "n0": [{"id": "MONDO:0000001", "attributes": []}],
    }


def test_batch_subclass(db_driver):
    """Test batched subclass query."""
    qgraph = {
        "nodes": {
            "n0": {"ids": [
                "MONDO:0000001",
                "HP:0000118",
            ]},
            "n1": {},
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 15
    for result in output["results"]:
        for binding in result["node_bindings"]["n0"]:
            assert binding["id"] in ["MONDO:0000001", "HP:0000118"]


def test_hierarchy_inference_on_superclass_queries(db_driver):
    """ Test that subclass edges should not be added to explicit subclass/superclass predicate queries  """
    qgraph = {
        "nodes": {"n0": {"ids": ["MONDO:0000001"]},
                  "n1": {}},
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "predicates": ['biolink:superclass_of']
            },
        }
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 2
    node_binding_ids = get_node_binding_ids_from_neo4j_driver_output(output)
    assert "MONDO:0000001" in node_binding_ids
    assert "MONDO:0005148" in node_binding_ids
    assert "MONDO:0015967" in node_binding_ids
    # note that for graphs where redundant subclass edges are added results would show MONDO:0000001 as a superclass
    # of MONDO:0014488, but because that explicit edge is not included in the test data MONDO:0014488 is excluded from
    # the results, as it's a subclass of a subclass of MONDO:0000001
    assert "MONDO:0014488" not in node_binding_ids


def test_hierarchy_inference_on_subclass_queries(db_driver):
    qgraph = {
        "nodes": {"n0": {"ids": ["MONDO:0005148"]},
                  "n1": {}},
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "predicates": ['biolink:subclass_of']
            },
        }
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 1
    node_binding_ids = [binding["id"] for binding_list in output["results"][0]["node_bindings"].values() for binding in
                        binding_list]
    assert "MONDO:0000001" in node_binding_ids
    assert "MONDO:0005148" in node_binding_ids
    assert "MONDO:0014488" not in node_binding_ids


def test_subclass_depth_default(db_driver):
    """Test one-hop subclass query."""
    qgraph = {
        "nodes": {"n0": {"ids": ["CHEBI:136043"]},
                  "n1": {"ids": ["MONDO:0000000"]}},
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "predicates": ['biolink:treats']
            },
        }
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 0


def test_subclass_depth_2(db_driver):
    """Test one-hop subclass query."""
    qgraph = {
        "nodes": {"n0": {"ids": ["CHEBI:136043"]},
                  "n1": {"ids": ["MONDO:0000000"]}},
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "predicates": ['biolink:treats']
            },
        }
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, subclass_depth=2, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 1


def test_invalid_subclass_depth(db_driver):
    qgraph = {
        "nodes": {"n0": {"ids": ["CHEBI:136043"]},
                  "n1": {"ids": ["MONDO:0000000"]}},
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "predicates": ['biolink:treats']
            },
        }
    }
    dialect, driver = db_driver
    with pytest.raises(TypeError):
        query = get_query(qgraph, subclass_depth="bad_value_type", dialect=dialect)
    with pytest.raises(ValueError):
        query = get_query(qgraph, subclass_depth=-1, dialect=dialect)



"""Test Python obj -> cypher expression encoding."""
from reasoner_transpiler.cypher_expression import dumps


def test_types():
    """Test various Python types."""
    # number
    assert dumps(1.2) == "1.2"
    # string
    assert dumps("hello") == "\"hello\""
    # list
    assert dumps(["hello", 1.2]) == "[\"hello\", 1.2]"
    # boolean
    assert dumps(True) == "true"
    # None
    assert dumps(None) == "null"
    # dict
    assert dumps({
        "a": 1,
        "b": False,
    }) == "{a: 1, b: false}"



import pytest

from reasoner_transpiler.cypher import get_query
from .fixtures import fixture_db_driver


def test_single_qualifier(db_driver):
    """Test edge satisfying one set of qualifier constraints is returned """
    qgraph = {
        "nodes": {
            "n0": {},
            "n1": {
                "ids": ["NCBIGene:283871"]
            },
        },
        "edges": {
            "e10a": {
                "subject": "n0",
                "object": "n1",
                "predicates": "biolink:affects",
                "qualifier_constraints": [
                    {
                        "qualifier_set": [
                            {
                                "qualifier_type_id": "biolink:qualified_predicate",
                                "qualifier_value": "biolink:causes"
                            }, {
                                "qualifier_type_id": "biolink:object_aspect_qualifier",
                                "qualifier_value": "activity"
                            }
                        ]
                    }
                ]
            },
        },
    }
    dialect,driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1
    assert len(output["results"][0]["analyses"][0]["edge_bindings"]["e10a"]) == 1
    qualified_predicate_output = {'qualifier_type_id': 'biolink:qualified_predicate', 'qualifier_value': 'biolink:causes'}
    qualifier_output = output["knowledge_graph"]["edges"]["qualified_edge_multiple_qualifier"]["qualifiers"]
    assert qualified_predicate_output in qualifier_output
    assert "qualified_predicate" not in output["knowledge_graph"]["edges"]["qualified_edge_multiple_qualifier"]
    assert "biolink:qualified_predicate" not in output["knowledge_graph"]["edges"]["qualified_edge_multiple_qualifier"]


def test_multi_qualifier(db_driver):
    """Test if edges satisfying constraints are returned with multiple qualifier sets"""
    qgraph = {
        "nodes": {
            "n0": {},
            "n1": {
                "ids": ["NCBIGene:283871"]
            },
        },
        "edges": {
            "e10a": {
                "subject": "n0",
                "object": "n1",
                "predicates": "biolink:affects",
                "qualifier_constraints": [
                    {
                        "qualifier_set": [
                            {
                                "qualifier_type_id": "biolink:object_aspect_qualifier",
                                "qualifier_value": "activity"
                            },
                        ]
                    },{
                        "qualifier_set": [
                            {
                                "qualifier_type_id": "biolink:qualified_predicate",
                                "qualifier_value": "biolink:causes"
                            }
                        ]
                    }
                ]
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1
    assert len(output["results"][0]["analyses"][0]["edge_bindings"]["e10a"]) == 2
    direction_qualifier_output = {'qualifier_type_id': 'biolink:object_direction_qualifier', 'qualifier_value': 'decreased'}
    qualified_predicate_output = {'qualifier_type_id': 'biolink:qualified_predicate', 'qualifier_value': 'biolink:causes'}
    aspect_qualifier_output = {'qualifier_type_id': 'biolink:object_aspect_qualifier', 'qualifier_value': 'activity'}
    qualifier_output = output["knowledge_graph"]["edges"]["qualified_edge_multiple_qualifier"]["qualifiers"]
    assert direction_qualifier_output in qualifier_output
    assert aspect_qualifier_output in qualifier_output
    assert qualified_predicate_output in qualifier_output


def test_qualifier_heirarchy(db_driver):
    """Test if edges satisfying constraints are returned when a qualifier has a value
    that is an ancestor of the value expected in the test data edge"""
    qgraph = {
        "nodes": {
            "n0": {},
            "n1": {
                "ids": ["NCBIGene:283871"]
            },
        },
        "edges": {
            "e10a": {
                "subject": "n0",
                "object": "n1",
                "predicates": "biolink:affects",
                "qualifier_constraints": [
                    {
                        "qualifier_set": [
                            {
                                "qualifier_type_id": "biolink:object_aspect_qualifier",
                                "qualifier_value": "activity_or_abundance"
                            },
                        ]
                    }
                ]
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1
    assert len(output["results"][0]["analyses"][0]["edge_bindings"]["e10a"]) == 1


def test_mismatch_qualifier_value(db_driver):
    """ test that no edges are returned with a qualifier value that doesn't match any edges"""
    qgraph = {
        "nodes": {
            "n0": {},
            "n1": {
                "ids": ["NCBIGene:283871"]
            },
        },
        "edges": {
            "e10a": {
                "subject": "n0",
                "object": "n1",
                "predicates": "biolink:affects",
                "qualifier_constraints": [
                    {
                        "qualifier_set": [
                            {
                                "qualifier_type_id": "biolink:object_aspect_qualifier",
                                "qualifier_value": "abundance"
                            },
                        ]
                    },
                ]
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 0


def test_empty_qualifier_set(db_driver):
    """Test if edges satisfying all constraints are returned"""
    qgraph = {
        "nodes": {
            "n0": {},
            "n1": {
                "ids": ["NCBIGene:283871"]
            },
        },
        "edges": {
            "e10a": {
                "subject": "n0",
                "object": "n1",
                "predicates": "biolink:affects",
                "qualifier_constraints": [
                    {
                        "qualifier_set": [
                        ]
                    },
                ]
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1



"""Test predicate handling."""
import pytest

from reasoner_transpiler.cypher import get_query
from .fixtures import fixture_db_driver


def test_symmetric(db_driver):
    """Test symmetric predicate."""
    qgraph = {
        "nodes": {
            "n0": {},
            "n1": {
                "ids": ["NCBIGene:836"]
            },
        },
        "edges": {
            "e10a": {
                "subject": "n1",
                "object": "n0",
                "predicates": "biolink:genetically_associated_with",
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 2


def test_any(db_driver):
    """Test any predicate."""
    qgraph = {
        "nodes": {
            "n0": {},
            "n1": {
                "ids": ["NCBIGene:836"]
            },
        },
        "edges": {
            "e10a": {
                "subject": "n1",
                "object": "n0",
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 4

def test_root_predicate(db_driver):
    """Test root/related_to predicate."""
    qgraph = {
        "nodes": {
            "n0": {},
            "n1": {
                "ids": ["NCBIGene:836"]
            },
        },
        "edges": {
            "e10a": {
                "subject": "n1",
                "object": "n0",
                "predicates": "biolink:related_to"
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 4

def test_sub(db_driver):
    """Test sub predicate."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["MONDO:0004993"]
            },
            "n1": {},
        },
        "edges": {
            "e10": {
                "subject": "n0",
                "object": "n1",
                "predicates": "biolink:genetically_associated_with",
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph,dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 2


def test_inverse(db_driver):
    """Test inverse predicate."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["NCBIGene:672"]
            },
            "n1": {},
        },
        "edges": {
            "e10": {
                "subject": "n0",
                "object": "n1",
                "predicates": "biolink:gene_associated_with_condition",
            },
        },
    }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph,dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1



import pytest

from reasoner_transpiler.cypher import get_query
from .fixtures import fixture_db_driver


def test_primary_source(db_driver):
    qgraph = {
            "nodes": {
                "n0": {
                    "categories": ["biolink:ChemicalSubstance"]
                },
                "n1": {
                    "categories": ["biolink:Disease"],
                    "ids": ["MONDO:0005148"]
                }

            },
            "edges": {
                "e0": {
                    "subject": "n0",
                    "object": "n1",
                    "predicate": "biolink:treats"
                }
            }
        }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 3
    assert len(output["knowledge_graph"]["edges"]) == 3

    edge_sources = {
        e: edge["sources"] for e, edge in output["knowledge_graph"]["edges"].items()
    }
    assert not any(['primary_knowledge_source' in edge for edge in output["knowledge_graph"]["edges"].values()])
    assert edge_sources["metformin_treats_t2d"] == [
        {'resource_id': 'infores:test', 'resource_role': 'primary_knowledge_source'},
        {'resource_id': 'infores:ctd', 'resource_role': 'aggregator_knowledge_source',
         'upstream_resource_ids': ['infores:test']},
        {'resource_id': 'reasoner-transpiler', 'resource_role': 'aggregator_knowledge_source',
         'upstream_resource_ids': ['infores:ctd']}]
    assert edge_sources["bezafibrate_treats_t2d"] == [
        {'resource_id': 'infores:test', 'resource_role': 'primary_knowledge_source'},
        {'resource_id': 'reasoner-transpiler', 'resource_role': 'aggregator_knowledge_source',
         'upstream_resource_ids': ['infores:test']}]
    assert edge_sources["anagliptin_treats_t2d"] == [
        {'resource_id': 'infores:test2', 'resource_role': 'primary_knowledge_source'},
        {'resource_id': 'reasoner-transpiler', 'resource_role': 'aggregator_knowledge_source',
         'upstream_resource_ids': ['infores:test2']}]


# if there is no primary knowledge source on the edge, provenance will be just the transpiler/plater
def test_missing_primary_source(db_driver):
    qgraph = {
            "nodes": {
                "n0": {
                    "ids": ["MESH:D014612"]
                },
                "n1": {
                    "categories": ["biolink:Disease"],
                    "ids": ["MONDO:0005260"]
                }

            },
            "edges": {
                "e0": {
                    "subject": "n0",
                    "object": "n1",
                    "predicate": "biolink:causes"
                }
            }
        }
    dialect, driver = db_driver
    output = driver.run(get_query(qgraph, dialect=dialect), convert_to_trapi=True, qgraph=qgraph)
    edge_sources = output["knowledge_graph"]["edges"]["invalid_provenance"]["sources"]
    assert edge_sources == [{'resource_id': 'reasoner-transpiler', 'resource_role': 'primary_knowledge_source'}]



"""Tools for compiling QGraph into Cypher query."""
import json

from collections import defaultdict

from .attributes import transform_attributes, PROVENANCE_TAG
from .matching import match_query


def nest_op(operator, *args):
    """Generate a nested set of operations from a flat expression."""
    if len(args) > 2:
        return [operator, args[0], nest_op(operator, *args[1:])]
    else:
        return [operator, *args]


def assemble_results(qnodes, qedges, **kwargs):
    """Assemble results into Reasoner format."""
    dialect = kwargs.get('dialect','neo4j')
    if dialect == 'memgraph':
        id_function = "id"
    elif dialect == 'neo4j':
        id_function = "elementId"


    clauses = []

    for qnode in qnodes.values():
        if qnode.get("set_interpretation", "") == "MANY":
            raise NotImplementedError(f'This feature is currently not implemented: set_interpretation=MANY')

    # TODO - implement pagination (SKIP and LIMIT)
    #  right now SKIP and LIMIT are unsupported and this will throw NotImplementedError if they are requested
    pagination(**kwargs)

    nodes = [f"`{qnode_id}`.id" for qnode_id, qnode in qnodes.items()]
    edges = [f"{id_function}(`{qedge_id}`)" if not qedge.get('_subclass', False)
             #else f"[x in `{qedge_id}` | {id_function}(x)]"
             else (f" CASE WHEN size(`{qedge_id}`) = 1 THEN [{id_function}(head(`{qedge_id}`))] ELSE [] END ")
             for qedge_id, qedge in qedges.items()]
    if nodes or edges:
        nodes_assemble = " + ".join([
            f"collect(`{qnode_id}`)"
            for qnode_id, qnode in qnodes.items()
        ])
        if not nodes_assemble:
            nodes_assemble = '[]'
        edges_assemble = " + ".join([
            f"collect([{id_function}(`{qedge_id}`), startNode(`{qedge_id}`).id, "
            f"type(`{qedge_id}`), endNode(`{qedge_id}`).id, properties(`{qedge_id}`)])"
            if not qedge.get('_subclass', False) else
            #f"collect([x in `{qedge_id}` | [{id_function}(x), startNode(x).id, "
            #f"type(x), endNode(x).id, properties(x)]])"
            f"collect( CASE WHEN size(`{qedge_id}`)= 1 THEN [ [{id_function}(head(`{qedge_id}`)), startNode(head(`{qedge_id}`)).id, "
            f"type(head(`{qedge_id}`)), endNode(head(`{qedge_id}`)).id, properties(head(`{qedge_id}`))]] ELSE [ ] END)"
            for qedge_id, qedge in qedges.items()
        ])
        if not edges_assemble:
            edges_assemble = '[]'
        assemble_clause = f"WITH {nodes_assemble} AS raw_nodes, " \
                          f"{edges_assemble} AS raw_edges, collect(DISTINCT ["
        if nodes:
            assemble_clause += ', '.join(nodes)
            if edges:
                assemble_clause += ', '
        if edges:
            assemble_clause += ', '.join(edges)
        assemble_clause += "]) AS paths "
        assemble_clause += "CALL { WITH raw_nodes UNWIND raw_nodes AS node RETURN collect(DISTINCT node) AS nodes } "
        assemble_clause += "CALL { WITH raw_edges UNWIND raw_edges AS edge RETURN collect(DISTINCT edge) AS edges } "
        clauses.append(assemble_clause)
        return_clause = "RETURN nodes, edges, paths"
    else:
        return_clause = 'RETURN [] as nodes, [] as edges, [] as paths'

    clauses.append(return_clause)
    return clauses


def pagination(skip=None, limit=None, **kwargs):
    """Get pagination clauses."""
    clauses = []
    if skip is not None:
        raise NotImplementedError(f'SKIP is not currently implemented.')
        # clauses.append(f"SKIP {skip}")
    if limit is not None:
        raise NotImplementedError(f'LIMIT is not currently implemented.')
        # clauses.append(f"LIMIT {limit}")
    return clauses


def get_query(qgraph, **kwargs):
    """Generate a Cypher query to extract the answer maps for a question.

    Returns the query as a string.
    """
    # commented this out because now we rely on the altering the qgraph to transform results into TRAPI,
    # leaving as a reminder in case that breaks something
    # qgraph = copy.deepcopy(qgraph)
    clauses = []
    query = match_query(qgraph, **kwargs)
    clauses.extend(query.compile())
    where_clause = query.where_clause()
    if where_clause:
        if not clauses[-1].startswith("WITH"):
            clauses.append(query.with_clause())
        clauses.append(where_clause)

    if not kwargs.pop("reasoner", True):
        clauses.append(query.return_clause())
        # add SKIP and LIMIT sub-clauses
        clauses.extend(pagination(**kwargs))
    else:
        clauses.extend(assemble_results(
            query.qgraph["nodes"],
            query.qgraph["edges"],
            **kwargs,
        ))

    return " ".join(clauses)


def transform_result(cypher_record,
                     qgraph: dict):

    nodes, edges, paths = unpack_bolt_record(cypher_record)

    # Construct the knowledge_graph["nodes"] section of the TRAPI response
    kg_nodes = {}
    for cypher_node in nodes:
        # Convert the list of unique result nodes from cypher results to dictionaries
        node = convert_bolt_node_to_dict(cypher_node)
        # Convert nodes to TRAPI format
        # id, name, and labels are removed before transform_attributes
        node_id = node.pop('id')
        kg_nodes[node_id] = {
            'name': node.pop('name'),
            'categories': sorted(node.pop('labels'))}
        kg_nodes[node_id].update(**transform_attributes(node, node=True))

    # Convert the list of unique edges from cypher results to dictionaries
    # then convert them to TRAPI format, constructing the knowledge_graph["edges"] section of the TRAPI response.
    # Also make a mapping of the neo4j element_id to the edge id to be used in the TRAPI edge bindings
    # the edge id used in TRAPI is the 'id' property on the edge if there is one, otherwise assigned integers 0,1,2..
    kg_edges, element_id_to_edge_id = transform_edges_list(edges)

    results = {}  # results are grouped by unique sets of result node ids
    aux_graphs = {}  # auxiliary_graphs

    qgraph_nodes = qgraph["nodes"]
    qnodes_with_set_interpretation_all = {qnode_id for qnode_id, qnode in qgraph_nodes.items()
                                          if qnode.get('set_interpretation', 'BATCH') == 'ALL'}
    qnodes_with_superclass_nodes = {qnode_id for qnode_id, qnode in qgraph_nodes.items()
                                    if f'{qnode_id}_superclass' in qgraph_nodes}

    qgraph_edges = qgraph["edges"]
    qedges_with_attached_subclass_edges = defaultdict(list)
    for qedge_id, qedge in qgraph_edges.items():
        if not qedge.get('_subclass', False):
            if qedge["subject"] in qnodes_with_superclass_nodes:
                qedges_with_attached_subclass_edges[qedge_id].append(
                    ('subject', f'{qedge["subject"]}_subclass_edge', f'{qedge["subject"]}_superclass'))
            if qedge["object"] in qnodes_with_superclass_nodes:
                qedges_with_attached_subclass_edges[qedge_id].append(
                    ('object', f'{qedge["object"]}_subclass_edge',  f'{qedge["object"]}_superclass'))

    # Each path is an array of nodes and edges like [n1, n2, n3, e1, e2, e3],
    # where nodes are node_ids from the graph and edges are element_ids of relationships from the graph.
    for path in paths:
        # Map results/paths to their corresponding qnodes and qedges
        qnode_id_to_results = {qnode_id: (qnode, result_node_id) for (qnode_id, qnode), result_node_id in
                               zip(qgraph_nodes.items(), path[:len(qgraph_nodes)])}
        qedge_id_to_results = {qedge_id: (qedge, result_edge_id) for (qedge_id, qedge), result_edge_id in
                               zip(qgraph_edges.items(), path[-len(qgraph_edges):])}

        # results are grouped by unique sets of nodes, concatenating node ids to this string generates the key
        result_key = ""

        # create TRAPI node bindings
        edge_bindings = {}
        node_bindings = {}
        for qnode_id, (qnode, result_node_id) in qnode_id_to_results.items():

            # don't return superclass qnodes in the node bindings
            if qnode.get('_superclass', False):
                continue
            # if there isn't a result for this node set an empty node binding
            if not result_node_id:
                node_bindings[qnode_id] = []
                continue

            # create a node binding
            if qnode_id in qnodes_with_set_interpretation_all:
                # if qnode has set_interpretation=ALL there won't be any superclass bindings
                node_bindings[qnode_id] = [{'id': result_node_id, 'attributes': []}]

            elif qnode_id in qnodes_with_superclass_nodes:
                # If the qnode has a superclass node, and it has a different result node id,
                # use the superclass result node id instead, because it's the actual query id.
                # We used to include the query_id property here to show that, but now we're making a support graph
                # that can represent the underlying subclass edge(s).
                superclass_qnode, superclass_result_id = qnode_id_to_results[f'{qnode_id}_superclass']
                node_bindings[qnode_id] = \
                    [{'id': result_node_id if superclass_result_id == result_node_id else superclass_result_id,
                      'attributes': []}]
            else:
                # Otherwise, create a normal node binding.
                node_bindings[qnode_id] = \
                    [{'id': result_node_id,
                      'attributes': []}]

            # add the result node id to the result key
            result_key += node_bindings[qnode_id][0]['id']

        # Create TRAPI edge bindings
        for qedge_id, (qedge, path_edge) in qedge_id_to_results.items():

            # skip empty results
            if ((not path_edge) and (not path_edge == 0)):
                continue
            # don't return subclass qedges in the edge bindings
            if qedge.get("_subclass", False):
                continue

            # find the knowledge graph edge id for the element id from the path edge
            edge_element_id = path_edge
            graph_edge_id = element_id_to_edge_id[edge_element_id]

            # Check to see if the edge has subclass edges that are connected to it
            subclass_edge_ids = []
            superclass_node_ids = {}
            for (subclass_subject_or_object, subclass_qedge_id, superclass_qnode_id) in \
                    qedges_with_attached_subclass_edges.get(qedge_id, []):
                # If so, check to see if there are results for it
                qedge, subclass_edge_element_ids = qedge_id_to_results[subclass_qedge_id]
                qedge, subclass_edge_element_ids = qedge_id_to_results[subclass_qedge_id]
                if subclass_edge_element_ids:
                    # If path_edge is Truthy, it means the subclass was used in the result.
                    # For subclass edges, path result is a list of element ids, due to being a variable length edge.
                    # make a list of the subclass edges plus the result edge from the query.
                    subclass_edge_ids.extend([element_id_to_edge_id[ele_id] for ele_id in subclass_edge_element_ids])

                    qnode, superclass_result_node_id = qnode_id_to_results[superclass_qnode_id]
                    superclass_node_ids[subclass_subject_or_object] = superclass_result_node_id

            if subclass_edge_ids:
                # make a composite id with all of their kg edge ids
                composite_edge_ids = [graph_edge_id] + subclass_edge_ids
                composite_edge_id = "_".join(composite_edge_ids)
                aux_graph_id = f"aux_{composite_edge_id}"
                if aux_graph_id not in aux_graphs:
                    aux_graphs[aux_graph_id] = {
                        "edges": composite_edge_ids,
                        "attributes": []
                    }
                if composite_edge_id not in kg_edges:
                    real_edge = kg_edges[graph_edge_id]
                    inferred_result_edge = {"subject": real_edge["subject"],
                                            "predicate": real_edge["predicate"],
                                            "object": real_edge["object"],
                                            "attributes": [
                                                {
                                                    "attribute_type_id": "biolink:knowledge_level",
                                                    "value": "logical_entailment"
                                                },
                                                {
                                                    "attribute_type_id": "biolink:agent_type",
                                                    "value": "automated_agent",
                                                },
                                                {
                                                    "attribute_type_id": "biolink:support_graphs",
                                                    "value": [aux_graph_id]
                                                }
                                            ],
                                            "sources": [
                                                {
                                                    "resource_id": PROVENANCE_TAG,
                                                    "resource_role": "primary_knowledge_source"
                                                }
                                            ],
                                            # this overwrites subject and/or object with the superclass node ids
                                            # if they exist
                                            **superclass_node_ids}
                    kg_edges[composite_edge_id] = inferred_result_edge

                # make an edge binding with the inferred subclass edge
                edge_bindings[qedge_id] = [{'id': composite_edge_id, 'attributes': []}]
            else:
                # if no subclass edges for this edge make a normal edge binding
                edge_bindings[qedge_id] = [{'id': graph_edge_id, 'attributes': []}]

        if result_key != '':  # avoid adding results for the default node binding key ''
            # if we haven't encountered this specific group of result nodes before, create a new result
            if result_key not in results:
                results[result_key] = {'node_bindings': node_bindings,
                                       'analyses': [{'edge_bindings': edge_bindings,
                                                     'resource_id': PROVENANCE_TAG}]}
            else:
                # otherwise append new edge bindings to the existing result
                for qedge_id, edge_binding_list in edge_bindings.items():
                    results[result_key]['analyses'][0]['edge_bindings'][qedge_id].extend(
                        [new_edge_bind for new_edge_bind in edge_binding_list if new_edge_bind['id'] not in
                         [existing_edge_bind['id'] for existing_edge_bind in
                          results[result_key]['analyses'][0]['edge_bindings'][qedge_id]]])

    knowledge_graph = {
        'nodes': kg_nodes,
        'edges': kg_edges
    }
    transformed_results = {
        'results': list(results.values()),  # convert the results dictionary to a flattened list
        'knowledge_graph': knowledge_graph,
        'auxiliary_graphs': aux_graphs
    }
    return transformed_results


def transform_edges_list(edges):
    # See convert_bolt_edge_to_dict() for details on the contents of edges,
    # it is a list of lists (which can also be lists), representing unique edges from the graph
    kg_edges = {}
    element_id_to_edge_id = {}
    edge_index = 1
    for cypher_edge_result in edges:
        # skip an empty list
        if len(cypher_edge_result) == 0:
            cypher_edges = []
        # check to see if it's a list of lists (multiple edges)
        elif isinstance(cypher_edge_result[0], list):
            cypher_edges = cypher_edge_result
        else:
            # otherwise it's just one list (one edge)
            cypher_edges = list()  # this looks weird, but it's necessary to make a list of one list (I think?)
            cypher_edges.append(cypher_edge_result)

        # transform the edge(s) into TRAPI
        for cypher_edge in cypher_edges:
            # get the element id from the cypher results (see convert_bolt_edge_to_dict for more details)
            edge_element_id = cypher_edge[0]
            # skip the edge if we already processed it
            # this prevents duplicating effort for cases where a subclass edge is included multiple times in a result
            # this can happen because neo4j tries to deduplicate, but variable length subclass edges fail
            # to match a single edge even when the single edge is part of the variable length edge
            if edge_element_id in element_id_to_edge_id:
                continue

            # transform the edge into TRAPI and return:
            # edge_id - the edge id that will be used for edges in the TRAPI knowledge graph and edge bindings
            # trapi_edge - a dictionary that represents an edge in the knowledge_graph part of the TRAPI response
            edge_id, trapi_edge = convert_bolt_edge_to_trapi(cypher_edge)
            if not edge_id:
                edge_id = f'e_{edge_index}'
            # make a mapping that will be used to look up the edge id by element id later
            element_id_to_edge_id[edge_element_id] = edge_id
            # add it to the knowledge graph
            kg_edges[edge_id] = trapi_edge
            edge_index += 1
    # returns the TRAPI knowledge graph edges and a lookup mapping of edge element id -> TRAPI edge id
    return kg_edges, element_id_to_edge_id


def convert_bolt_node_to_dict(bolt_node):
    if not bolt_node:
        return None
    node = {key: value for key, value in bolt_node.items()}
    node['labels'] = bolt_node.labels
    return node


def convert_jolt_node_to_dict(jolt_node):
    if not jolt_node:
        return None
    node = jolt_node['()'][2]
    node['element_id'] = jolt_node['()'][0]
    node['labels'] = jolt_node['()'][1]
    return node


# Convert a list representing an edge from cypher results into a dictionary.
# This is not any standard object from the bolt driver, it's a list generated by a specific cypher return clause, like:
# [elementId(edge_1), startNode(edge_1).id, type(edge_1), endNode(edge_1).id, properties(edge_1)]
# This is done to prevent including often redundant node and edge properties on nodes and edges in pathway results.
# See the cypher generated in the edges_assemble clause in assemble_results() for more details.
def convert_bolt_edge_to_trapi(bolt_edge):
    if not bolt_edge:
        print(f'Tried to convert a missing edge: {bolt_edge}')
        return None, None

    converted_edge = {
        'subject': bolt_edge[1],
        'predicate': bolt_edge[2],
        'object': bolt_edge[3],
    }
    # edge_props - any other properties from the edge
    edge_props = {**bolt_edge[4]}

    # retrieve and remove the id if there is one on the edge
    edge_id = edge_props.pop('id', None)

    # convert all remaining attributes to TRAPI format, constructing the attributes and sources sections
    converted_edge.update(transform_attributes(edge_props, node=False))

    # return the edge id if there was one, and a TRAPI edge
    return edge_id, converted_edge


def convert_jolt_edge_to_dict(jolt_edges, jolt_element_id_lookup):
    if not jolt_edges:
        return None
    convert_edges = []
    jolt_edges = [jolt_edges] if not isinstance(jolt_edges, list) else jolt_edges
    for jolt_edge in jolt_edges:
        jolt_edge = next(iter(jolt_edge.values()))
        edge = jolt_edge[4]  # start with the properties
        # add the SPO from the other items of the jolt edge list
        edge.update({
            'subject': jolt_element_id_lookup[jolt_edge[1]],
            'predicate': jolt_edge[2],
            'object': jolt_element_id_lookup[jolt_edge[3]]})
        if 'id' not in edge:
            edge['id'] = jolt_edge[0]  # if the edge didn't have an "id" use the neo4j element id
        convert_edges.append(edge)
    return convert_edges


def unpack_bolt_record(bolt_record):
    return bolt_record['nodes'], bolt_record['edges'], bolt_record['paths']


def unpack_jolt_result(jolt_response):
    headers = []
    for line in jolt_response.split("\n"):
        line = json.loads(line)
        if 'header' in line:
            headers = line['header']['fields']
        elif 'data' in line:
            data = {header: data_item for (header, data_item) in zip(headers, line['data'])}
            return data['nodes'], data['edges'], data['paths']
